#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "expr.h"
#include "widget.h"
#include "strbuf.h"
#include "eval.h"
#include "library.h"
#include "ptrbuf.h"
#include "gen.h"
#include "n-eval.h"

static char *dotless_name(char *str, char *pfx)
{
  char *Z;
  strbuf_t *SB;
  SB= strbuf_init();
  strbuf_puts(SB, pfx);
  strbuf_puts(SB, "_");
  while(1)
  {
    Z= strchr(str,'.');
    if (Z)
    {
      if (Z!=str)
        strbuf_puts_l(SB, str, Z-str);
      strbuf_puts(SB,"_XO_");
      str= Z+1;
      if (!str[0]) break;
    }
    else
    {
      strbuf_puts(SB,str);
      break;
    }
  }
  return strbuf_collect(SB);
}

static char *widget_layout_fname(Uscreen_t *scr)
{
  return dotless_name(scr->ident, "LAYOWID");
}

static char *screen_layout_fname(Uscreen_t *scr)
{
  return dotless_name(scr->ident,"LAYOSCR");
} 

static void gen_widget_layout(Uscreen_t *scr,FILE* f)
{
  char *code;
  size_t nunk;
  unknown_t **unk;
  int N;
  Uwidget_t *W;
  int n_locals;

  bc_buffer_t *nc;

  nc= compile_widget_layout(scr, &n_locals);
  if (nc)
  {
    print_bytecode("out.bc", nc);
  }
  else
  {
    printf("compilation failed\n");
    return ;
  }

  fprintf(f,"static const bytecode_t %s= {\n", widget_layout_fname(scr));
  code= bc_buffer_to_string(nc, n_locals, 0);
  fprintf(f,"%s",code); free(code);
  fprintf(f,"};\n");
  
  
#if 0
  fprintf(f,"static void %s(int WIDTH,int HEIGHT,int *L) {\n",
                  widget_layout_fname(scr));
  fprintf(f, "  int locals[%d];\n", n_locals);
  fprintf(f, "%s",
"locals[GUILV_WIDTH]= WIDTH;\n"
"locals[GUILV_HEIGHT]= HEIGHT;\n");
  fprintf(f,"static const unsigned char bytes[]= {");
  code= bc_buffer_to_string(nc, n_locals, 1);
  fprintf(f,"%s",code); free(code);
  fprintf(f,"};\n");

  fprintf(f,"static const bytecode_t code= {\n");
  fprintf(f," .length= %d, .n_locals= %d, .data= bytes };\n", nc->len, 
     n_locals);
  bc_buffer_free(nc);
  fprintf(f,"gui_run_bytecode(code.data, code.length, locals, L);\n");
  fprintf(f,"}\n");
#endif


#if 0

  fprintf(f,"static void %s(int WIDTH,int HEIGHT,int *L) {\n",
                  widget_layout_fname(scr));
  code= eval_widget_layout(scr, &nunk, &unk);
  if (nunk)
  { 
     int i;
     fprintf(f,"%s",code);
     fprintf(f,"#error \"there are unknowns, eval this first\"\n");
     for(i=0;i<nunk;i++)
      fprintf(f,"// %s= %s \n", unk[i]->target, unk[i]->str);
     fprintf(f,"}\n");
     return ;
  }
  fprintf(f,"%s",code);
  free(code);

  for(N=0,W=scr->widgets;W;W=W->next,N++)
  {
    fprintf(f, "  L[%d]= CO%s_left; L[%d]= CO%s_top;\n" 
               "  L[%d]= CO%s_width; L[%d]= CO%s_height;\n", 
                     N*4, W->ident, N*4+1, W->ident,
                     N*4+2, W->ident, N*4+3, W->ident);
  }
  fprintf(f,"}\n");
#endif
}

static void gen_screen_dim(Uscreen_t *scr, FILE *f)
{
  fprintf(f,"static void %s(int *L,int WIDTH,int HEIGHT) {\n",
                       screen_layout_fname(scr));
  fprintf(f,"%s", eval_screen_dim(scr));
  fprintf(f,"  L[0]= COleft;\n");
  fprintf(f,"  L[1]= COtop;\n");
  fprintf(f,"  L[2]= COwidth;\n");
  fprintf(f,"  L[3]= COheight;\n");
  fprintf(f,"}\n");
}

static void gen_widget_names(FILE *f)
{
  Uscreen_t *scr;
  Uwidget_t *W;
  Uscreen_t **SS; int N;
  int i;
  int last;

  SS= get_all_screens(&N);
     // Warning!
     // These are from gfx.h if you change those, you should
     // come here and change these also.
#ifndef GFXEV_LAST
#define GFXEV_LAST 14
#endif
#ifndef GFXRESERVED
#define GFXRESERVED 24
#endif

  fprintf(f,"enum { _XX_WIDGETS_START=GFXEV_LAST+GFXRESERVED,\n ");
  fprintf(f," /* WIDGET NAMES BELOW */\n");
  char buf[2000];
  buf[0]= 0;

  last= GFXEV_LAST+GFXRESERVED+1;
  for(i=0;i<N;i++)
  {
   scr= SS[i];
   for(W=scr->widgets;W;W=W->next)
   {
     fprintf(f,"  %s%s,\n", W->ident,buf);
     W->global_number= last;
     if (!strcmp(W->klass->ident,"PixGrid"))
     {
       sprintf(buf,"=%s+GFXEV_LAST",W->ident);
       last+= GFXEV_LAST;
     }
     else
     {
       last++;
       buf[0]= 0;
     }
   }
  }
  fprintf(f,"\n _XX_WIDGETS_END };\n");
  free(SS);
}

typedef struct { int len; FILE *file; } array_out_t;

static void AOs(array_out_t *AO, char *str)
{
  int L;
  L= strlen(str);
  if (AO->len+L>72) { fprintf(AO->file,"\n"); AO->len= 0; }
  AO->len+= L;
  fprintf(AO->file,"%s",str);
}

static void AOinit(array_out_t *AO, FILE *f, char *init)
{
  AO->file= f;
  AO->len= 0;
  AOs(AO,init);
}

static void AOi(array_out_t *AO, int V)
{
  char buf[20];
  sprintf(buf,"%d,",V);
  AOs(AO,buf);
}

static void declare_widget_list(Uwidget_t **WL, int n, FILE *f)
{
  int i;
  array_out_t A;
  AOinit(&A, f, " static int SWID[]= { ");
  for(i=0;i<n;i++) AOi(&A,WL[i]->global_number);
  fprintf(f,"};\n");
}

static Uwidget_t** get_widget_list_breadthfirst(Uscreen_t *scr,int *Rn)
{
  // It's important to get the list in breadthfirst fashion,
  // this is because we want to calculate the size&pos for the
  // parent widgets first, and then the child widgets. 
  ptrbuf_t *queue;
  ptrbuf_t *result;
  Uwidget_t *W;
  int i;
  size_t Nww; Uwidget_t **ww;
  int nresult; // #widgets in result so far

  result= ptrbuf_new(20); // holds the widgets
  queue= ptrbuf_new(20); // holds the screens
  nresult= 0;

  ptrbuf_put(queue, scr);
  while((scr=ptrbuf_get(queue)))
  {
    scr->windex_start= nresult;
    for(W=scr->widgets;W;W=W->next)
    {
      ptrbuf_put(result, W);
      nresult++;
      scr->windex_end= nresult;
      for(i=0;i<W->n_subs;i++) ptrbuf_put(queue, W->subs[i]);
    }
  }
  ptrbuf_free(queue,NULL,NULL);
  ptrbuf_free(result, (void***) &ww, &Nww);
  *Rn= Nww;
  return ww;
}

static void Gcreate_widget
  (Uscreen_t *scr,Uwidget_t *W,FILE *f, bc_buffer_t *bc)
{
  if (W->klass->construct)
  {
     (*W->klass->construct)(scr,W,bc);
  }
  else 
  {
     (*W->klass->create)(W,f);
     fprintf(f, "  widgets[%s]->number= %s;\n", W->ident, W->ident);
  }
}



static void gen_screen_func(Uscreen_t *scr,FILE *f,int mode)
{
  Uwidget_t *W;
  int i,j;
  Uscreen_t *S;
  bc_buffer_t *cons;
  char *str;

  Uwidget_t **ww; int Nww; // this is the list of ALL widgets,
                           // including those in the sub widgets.
  int Nw;
    // This is the number of widgets directly contained in
    // the screen. This number of widgets appear first in
    // the widgetlist, since it's a breadth-first list.
    // Nww and Nw are used differently. All widgets are
    // subjected to some operations such as translating
    // them by the origin of the screen. For widgets directly
    // contained in the screen, we do other things such as
    // showing them. The difference between these two stem
    // from panel widgets, which can contain other widgets.

  cons= bc_buffer_new(1000);


  fprintf(f,"void gfx_%s()",scr->ident);
  if (mode=='H') { fprintf(f,";\n"); return ; }
  fprintf(f,"{\n");

  ww= get_widget_list_breadthfirst(scr,&Nww);

  for(Nw=0,W=scr->widgets;W;W=W->next,Nw++)
    if (W!=ww[Nw]) break;

  fprintf(f, " int n_widgets= %d, n_direct_widgets= %d; \n", Nww, Nw);
  declare_widget_list(ww, Nww,f);
  fprintf(f,"  int i,j,start,end,pindex;\n");
  fprintf(f,"  int LC[%d]; /*widget coords */\n",Nww*4);
  fprintf(f,"  int LL[4]; /* screen coords */\n");

  fprintf(f,"  %s(LL,x_window_width,x_window_height);\n",
                   screen_layout_fname(scr));

  fprintf(f,"  enter_screen(%d, SWID,LL);\n",Nww);

                          // we create everything here, including
                          // the widgets embedded in panels.
  for(i=0;i<Nww;i++) Gcreate_widget(scr,ww[i],f, cons);

  fprintf(f,"  static const int cons_length= %d;\n", cons->len);
  fprintf(f,"  static const unsigned char cons_code[]= {\n");
  str= bc_buffer_to_string(cons, 0, 1);
  fprintf(f,"%s};\n", str);
  free(str);

  fprintf(f,"  gui_construct_widgets(cons_code, cons_length);\n");

  fprintf(f,"  gui_layout_widgets(LL[2],LL[3], &%s ,LC);\n", 
              widget_layout_fname(scr));

  // Now we need to layout the subscreens. We do that using the
  // breadth-first widget list, which gives us a breadth-first
  // sub-screen list. 

  int n_subs= 0;
  for(i=0;i<Nww;i++)
    if (ww[i]->n_subs) n_subs+= ww[i]->n_subs;

  fprintf(f,"int n_subscreens= %d;\n", n_subs);
  fprintf(f,"static const struct { \n");
  fprintf(f,"   int parent_idx, start_idx, end_idx;\n");
  fprintf(f,"   const bytecode_t *code;\n");
  fprintf(f,"} subscreens[%d]= {\n", n_subs);

  for(i=0;i<Nww;i++)
  {
    W= ww[i];
    if (!W->n_subs) continue;
    for(j=0;j<W->n_subs;j++)
    {
      S= W->subs[j];
      fprintf(f,"  { %d, %d, %d, &%s },\n", i, S->windex_start, S->windex_end,
                                               widget_layout_fname(S));
    }
  }
  fprintf(f,"};\n");

  fprintf(f,"for(i=0;i<n_subscreens;i++)\n");
  fprintf(f,"{\n");
  fprintf(f,"  pindex= subscreens[i].parent_idx;\n");
  fprintf(f,"  start= subscreens[i].start_idx;\n");
  fprintf(f,"  end= subscreens[i].end_idx;\n");
  fprintf(f,"  gui_layout_widgets(LC[pindex*4+2], LC[pindex*4+3],\n");
  fprintf(f,"                     subscreens[i].code, LC+4*start);\n");
  fprintf(f,"  for(j=start;j<end;j++)\n");
  fprintf(f,"  {\n");
  fprintf(f,"     LC[4*j]+= LC[4*pindex];\n");
  fprintf(f,"     LC[4*j+1] += LC[4*pindex+1];\n");
  fprintf(f,"  }\n");
  fprintf(f,"}\n");

  fprintf(f,"  for(i=0;i<n_widgets;i++)\n");
  fprintf(f,"  {\n");
  fprintf(f,"    LC[4*i]+= LL[0]; LC[4*i+1]+= LL[1];\n");
  fprintf(f,"    widget_move(widgets[SWID[i]], LC[4*i], LC[4*i+1],\n");
  fprintf(f,"                                  LC[4*i+2], LC[4*i+3]);\n");
  fprintf(f,"  }\n");

  fprintf(f,"  for(i=0;i<n_direct_widgets;i++)\n");
  fprintf(f,"     widget_show(widgets[SWID[i]],1);\n");
  fprintf(f,"  gfx_update_all();\n");
  fprintf(f,"}\n");
}

static char *replace_suffix(char *str,char *sfx)
{
  int i;
  str= strdup(str);
  for(i=strlen(str)-1;i>0;i--)
     if (str[i]=='.') { strcpy(str+i,sfx); break; }
  return  str;
}


static void gen_colorimg_resource(FILE *f,resource_t *R)
{
  int i;
  int L;
  fprintf(f,"RES_COLORIMG(%s,\"User resource\")",R->ident);
  L= R->length/4;
  for(i=0;i<L;i++)
  {
    if (i%5==0) fprintf(f,"\n");
    fprintf(f,"0x%08x,", *(uint32_t*) (R->data+4*i));
  }
  fprintf(f,"\nRES_END\n");
}

static void gen_resources(FILE *f)
{
  resource_t *R;
  for(R=resources;R;R=R->next)
  {
   switch(R->type)
   {
   case TYP_COLORIMG:
    gen_colorimg_resource(f,R);
    break;
   }
  }
}

static void gen_resource_iface(FILE *f)
{
  resource_t *R;
  for(R=resources;R;R=R->next)
  {
   switch(R->type)
   {
   case TYP_COLORIMG:
    fprintf(f, "extern unsigned int %s[];\n", R->ident);
    break;
   }
  }
}

static void generate_header(FILE *out)
{
  const char *Chdr;
  int i,N;
  Uscreen_t **SS,*S;

  get_library_code(&Chdr, NULL);

  fwrite(Chdr, 1, strlen(Chdr), out);

  gen_widget_names(out);
  SS= get_all_screens(&N);

  for(i=0;i<N;i++)
  {
    S= SS[i];
    if (S->type) continue;
    gen_screen_func(S, out,'H'); 
  }
  free(SS);
  gen_resource_iface(out);
}

const char *print_to_mark(const char *code, const char *mark, FILE *f)
{
  const char *z;
  z= strstr(code,mark);
  if (z)
    { while(*z!=0 && *z!='\n') z++;
      if (*z=='\n') z++; }
  else
    z= code+strlen(code);
  if (f)
    fwrite(code, 1, z-code,f);
  return z;
}


static void generate_source_file(char *fn)
{
  FILE *out;
  const char *Ccode;
  int i,N;
  Uscreen_t **SS,*S;

  out= fopen(fn, "w");
  if (!out) return ; // FIXME: give error message

  if (getenv("GFXGEN_DBG")) 
    Ccode= "";
  else
    get_library_code(NULL, &Ccode);

  Ccode= print_to_mark(Ccode, "/** INCLUDE DIRECTIVE BEGIN **/", out);
  generate_header(out);
  Ccode= print_to_mark(Ccode, "/** INCLUDE DIRECTIVE END **/", NULL);
  Ccode= print_to_mark(Ccode, "/** USER RESOURCES **/", out);
  gen_resources(out);
  fwrite(Ccode, 1, strlen(Ccode), out);

  SS= get_all_screens(&N);

  for(i=0;i<N;i++)
  {
    S= SS[i];
    if (!S->type) continue;
    gen_widget_layout(S, out);
  }
 
  for(i=0;i<N;i++)
  {
    S= SS[i];
    if (S->type) continue;
    gen_widget_layout(S, out);
    gen_screen_dim(S, out);
    gen_screen_func(S, out,'C'); 
  }
  free(SS);

  fclose(out);
}

void generate_code(char *hfilenm,char *cfilenm)
{
  FILE *out;
  if (cfilenm) generate_source_file(cfilenm);
  if (!hfilenm) return ;
  out= fopen(hfilenm,"w");
  if (!out) return ;
  generate_header(out);
  fclose(out);
}

/***********************************************************************
   The layout object.

   This object contains all the code necessary to lay out widgets in a
   screen. It has two parts: main screen widget layout followed by 
   subscreen layouts. The main screen layout part has the following
   parameters:
    n_widgets
    n_direct_widgets
    n_subscreens
    n_locals
    length_of_code
   this is followed by the actual code. 

   After the main part, we have a subscreen layout section for each
   subscreen. A subscreen layout section has the following parameters:
    parent_index
    start_index
    end_index
    n_locals
    length_of_code
   This is followed by the code for the subscreen layout.
 ***********************************************************************/

