#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "strbuf.h"
#include "ptrbuf.h"
#include "expr.h"
#include "widget.h"
#include "eval.h"

typedef struct
{
  char *name;
  int *sto;
} known_t;

static known_t *kno; static size_t nkno;

static int f_wid= 800, f_hei= 600;
static int fscw= 12, fsch= 12, f_sbar=25;
static int f_fh1= 18, f_fh2= 18, f_fh3= 18, f_fh4= 18, f_fh5= 18;
static known_t fixed[]=
{ 
  { "WIDTH", &f_wid },
  { "HEIGHT", &f_hei},
  { "gui_scrollbar_width", &fscw}, 
  { "gui_scrollbar_height", &fsch}, 
  { "gui_sysbar_height", &f_sbar },
  { "font1height", &f_fh1 },
  { "font2height", &f_fh2 },
  { "font3height", &f_fh3 },
  { "font4height", &f_fh4 },
  { "font5height", &f_fh5 },
};
static int nfixed= 10;

static char *Cnames[6]= 
        { "top", "bottom", "left", "right", "width", "height" };

static int evalbinop(expr_t *E, strbuf_t *buffer, char *op);
static int evalpreop(expr_t *E, strbuf_t *buffer, char *op);

static void copy_fixed_knowns(size_t nunk)
{
  int i;
  if (kno) free(kno); 
  kno= malloc(sizeof(*kno)*(nunk + nfixed));
  for(i=0;i<nfixed;i++) kno[i]= fixed[i];
  nkno= nfixed;
}

static void makeref
  (ptrbuf_t *B, Uwidget_t *W, int tgt, int s1, char *op, int s2) 
{
  char *str; expr_t* expr; 
  char nb[200];

  expr= W->pos.E[tgt];
  str= W->pos.S[tgt];
  if (!expr)
  {
    str= malloc(strlen(W->ident)*2+2*8+strlen(op)+10);
    sprintf(str,"CO%s_%s%sCO%s_%s",W->ident,Cnames[s1],op,W->ident,Cnames[s2]);
    expr= xpp_run(xpp, str);
    if (!expr || expr->op==XPP_ERROR) 
    {
      // this will never happen. in case it does, I want a very weird
      // but consistent output
      expr= xpp_run(xpp, "0");
    }
  }

  sprintf(nb,"CO%s_%s",W->ident,Cnames[tgt]);
  unknown_t *U;
  U= calloc(1,sizeof(*U));
  U->target= strdup(nb);
  U->expr= expr;
  U->str= str;

  U->source= 'W';
  U->ptr= W;
  U->subsource= tgt;

  switch(tgt)
  {
  case wTOP: U->store= &W->pos.Dy; break;
  case wBOT: U->store= &W->pos.Db; break;
  case wLEF: U->store= &W->pos.Dx; break;
  case wRIG: U->store= &W->pos.Dr; break;
  case wWID: U->store= &W->pos.Dw; break;
  case wHEI: U->store= &W->pos.Dh; break;
  }
  
  ptrbuf_put(B,U);
}

static void get_widget_layout_unknowns
   (Uscreen_t *S,unknown_t ***Runk,size_t *Rnunk)
{
  Uwidget_t *W;
  ptrbuf_t *Bunk;
  Uvar_t *V;

  Bunk= ptrbuf_new(50);

  for(W=S->widgets;W;W=W->next)
  {
    makeref(Bunk, W, wLEF, wRIG, "-", wWID);
    makeref(Bunk, W, wRIG, wLEF, "+", wWID);
    makeref(Bunk, W, wTOP, wBOT, "-", wHEI);
    makeref(Bunk, W, wBOT, wTOP, "+", wHEI);
    makeref(Bunk, W, wWID, wRIG, "-", wLEF);
    makeref(Bunk, W, wHEI, wBOT, "-", wTOP);
  }

  for(V=S->vars;V;V=V->next)
  {
    char buf[200];
    unknown_t *U;
    sprintf(buf,"V%s",V->ident);
    U= calloc(1,sizeof(*U));
    U->target= strdup(buf);
    U->str= V->str;
    U->expr= V->expr; 
    U->source= 'V';
    U->ptr= V;
    U->store= &V->sto;
    ptrbuf_put(Bunk, U);
  }
  ptrbuf_free(Bunk, (void***) Runk, Rnunk);

  copy_fixed_knowns(*Rnunk);
}

static int eval_sub(expr_t *E, strbuf_t *buffer)
{
  int i;
  switch(E->op)
  {
   case EXP_ADD: return evalbinop(E, buffer, "+");
   case EXP_SUB: return evalbinop(E, buffer, "-");
   case EXP_MUL: return evalbinop(E, buffer, "*");
   case EXP_DIV: return evalbinop(E, buffer, "/");
   case EXP_MOD: return evalbinop(E, buffer, "%");
   case EXP_NEG: return evalpreop(E, buffer, "-");
   case EXP_PLUS: return eval_sub(E->right, buffer);
   case EXP_PAREN_R:
     strbuf_puts(buffer,"(");
     if (eval_sub(E->left,buffer)) return 1;
     strbuf_puts(buffer,")");
     return 0;
   case XPP_FLO:
   case XPP_INT: strbuf_puts(buffer, E->data); return 0;
   case XPP_ERROR: return 1;
   case XPP_IDENT:
     for(i=0;i<nkno;i++)
     {
         if (kno[i].name[0]=='V' && !strcmp(kno[i].name+1,E->data))
         {
            strbuf_puts(buffer,kno[i].name); return 0; 
         }
         if (!strcmp(kno[i].name,E->data)) 
         {
           strbuf_puts(buffer, E->data); return 0; 
         }
     }
     break;
   case EXP_DOT:
     { expr_t *L,*R;
       L= E->left; R= E->right;
       if (L->op!=XPP_IDENT || R->op!=XPP_IDENT) return 1;
       Uwidget_t *W;
       W= find_widget(L->data); 
       if (!W) { 
         // fprintf(stderr,"bad widget: %s\n",L->data);
         return 1; 
       }
       for(i=0;i<6;i++) if (!strcmp(R->data,Cnames[i])) break;
       if (i==6) { 
         // fprintf(stderr,"bad field %s\n",R->data); 
         return 1;
       }
       char *Z;
       Z= malloc(2+strlen(W->ident)+1+strlen(R->data)+1);
       sprintf(Z,"CO%s_%s",W->ident,(char*) R->data);
       for(i=0;i<nkno;i++) if (!strcmp(kno[i].name, Z))
            {  strbuf_puts(buffer, Z); free(Z); return 0; }
       free(Z);
       return 1; }
  }
  return 1;
}

static int evalbinop(expr_t *E, strbuf_t *buffer, char *op)
{
  if (eval_sub(E->left, buffer)) return 1;
  strbuf_puts(buffer,op);
  if (eval_sub(E->right, buffer)) return 1;
  return 0;
}

static int eval_expr(char *target, expr_t *E, strbuf_t *buffer)
{
  int pos;
  pos= strbuf_pos(buffer);
  strbuf_printf(buffer," int %s= ", target);
  if (eval_sub(E,buffer))
  {
    strbuf_reset(buffer,pos);
    return 0;
  }
  strbuf_printf(buffer,";\n");
  return 1;
}



/** 
  Returns the  code necessary to execute the assignments.
  If there is already code in Bcode, the new code is appended
  to it. The buffer Bcode is deallocated by the end of this
  function.
 **/
static char *eval_assignments(strbuf_t *Bcode,size_t *Rnunk, unknown_t **unk)
{
  int i,change;
  size_t nunk;
  char *R;

  nunk= *Rnunk;

            // FIXME: What happens to stuff in Bcode here,
            // how does this function get called with non-NULL Bcode?
  if (!nunk) return strdup("//NOTHING\n"); 

  if (!Bcode) Bcode= strbuf_init();

  while(nunk)
  {
    change= 0;
    for(i=0;i<nunk;i++)
    {
      //printf("evaluating %s= %s .. ", unk[i]->target,unk[i]->str);
       if (eval_expr(unk[i]->target,unk[i]->expr, Bcode))
       {
         kno[nkno].name= unk[i]->target;
         kno[nkno].sto= unk[i]->store;
         nkno++;
         unk[i]= unk[nunk-1];
         nunk--;
         i--;
         change= 1;
       //printf(" OK\n");
       }
      //else printf("FAIL\n");
    }
    if (!change) break;
  } 

  R= strbuf_collect(Bcode);
  *Rnunk= nunk;
  return R;
}

static int evalpreop(expr_t *E, strbuf_t *buffer, char *op)
{
  strbuf_printf(buffer," %s",op);
  return eval_sub(E->left,buffer);
}


/****
For numeric evaluation, we should have some sort of a 'known' array, which
can be used to store the values. Hmm. I could have it like this:
  
 For each unknown, I can have a storage place. Then, I can record this
 storage place in the 'known' array and look it up when necessary. so,
 the variables will have some field for this. I will also need 
 Dbot and Drig for the widgets.. nice.

Alright, now I can do the evaluation thing.
****/


static int evaldis_binop(expr_t *E, double *RV, double (*f)(double,double));
static int evaldis_preop(expr_t *E, double *RV, double (*f)(double));

static double opADD(double A,double B) { return A+B; }
static double opSUB(double A,double B) { return A-B; }
static double opMUL(double A,double B) { return A*B; }
static double opDIV(double A,double B) { if (B!=0) return A/B; else return A; }
static double opMOD(double A,double B) { if ((int)B!=0)return (int)A%(int)B;
                                         else return A; }
static double opNEG(double A) { return -A; }


static int evaldis_sub(expr_t *E, double *RV)
{
  int i;
  switch(E->op)
  {
   case EXP_ADD: return evaldis_binop(E, RV, opADD);
   case EXP_SUB: return evaldis_binop(E, RV, opSUB);
   case EXP_MUL: return evaldis_binop(E, RV, opMUL);
   case EXP_DIV: return evaldis_binop(E, RV, opDIV);
   case EXP_MOD: return evaldis_binop(E, RV, opMOD);
   case EXP_NEG: return evaldis_preop(E, RV, opNEG);
   case EXP_PAREN_R:
   case EXP_PLUS: return evaldis_sub(E->left, RV);
     return 0;
   case XPP_FLO: *RV= strtod(E->data,NULL); return 0;
   case XPP_INT: *RV= atoi(E->data); return 0;
   case XPP_ERROR: return 1;
   case XPP_IDENT:
     for(i=0;i<nkno;i++)
     {
         if (kno[i].name[0]=='V' && !strcmp(kno[i].name+1,E->data))
         {
            *RV= *kno[i].sto;
            return 0;
         }
         if (!strcmp(kno[i].name,E->data)) 
         {
            *RV= *kno[i].sto;
            return 0;
         }
     }
     break;
   case EXP_DOT:
     { expr_t *L,*R;
       L= E->left; R= E->right;
       if (L->op!=XPP_IDENT || R->op!=XPP_IDENT) return 1;
       Uwidget_t *W;
       W= find_widget(L->data);
       if (!W) return 1;
       for(i=0;i<6;i++) if (!strcmp(R->data,Cnames[i])) break;
       if (i==6) return 1;
       char *Z;
       Z= malloc(2+strlen(W->ident)+1+strlen(R->data)+1);
       sprintf(Z,"CO%s_%s",W->ident,(char*) R->data);
       for(i=0;i<nkno;i++) if (!strcmp(kno[i].name, Z))
            {  *RV= *kno[i].sto; free(Z); return 0; }
       free(Z);
       return 1; }
  }
  return 1;
}


static int evaldis_expr(int *STO, expr_t *E)
{
  double D;
  if (evaldis_sub(E, &D)) return 0;
  if (STO) *STO= D;
  return 1;
}

static int evaldis_binop(expr_t *E, double *RV, double (*f)(double,double))
{
  double A,B;
  if (evaldis_sub(E->left, &A)) return 1;
  if (evaldis_sub(E->right, &B)) return 1;
  *RV= (*f)(A, B);
  return 0;
}


static int evaldis_preop(expr_t *E, double *RV, double (*f)(double))
{
  double A;
  if (evaldis_sub(E->left, &A)) return 1;
  *RV= (*f)(A);
  return 0;
}


void evaldis_coords(Uscreen_t *S,int sw,int sh)
{
  int i,change;
  unknown_t **unk; size_t nunk;

  f_wid= sw;
  f_hei= sh;

  get_widget_layout_unknowns(S, &unk, &nunk);

  if (!nunk) return ;

  while(nunk)
  {
    change= 0;
    for(i=0;i<nunk;i++)
    {
       if (evaldis_expr(unk[i]->store,unk[i]->expr))
       {
         kno[nkno].name= unk[i]->target;
         kno[nkno].sto= unk[i]->store;
         nkno++;
         unk[i]= unk[nunk-1];
         nunk--;
         i--;
         change= 1;
       }
    }
    if (!change) break;
  } 
  if (unk) free(unk);
}
  

static void free_unknown(unknown_t *U)
{
  if (U) free(U);
}

static void makeSref
  (ptrbuf_t *B, Uscreen_t *S, int tgt, int s1, char *op, int s2) 
{
  char *str; expr_t* expr; 
  char nb[200];

  expr= S->pos.E[tgt];
  str= S->pos.S[tgt];
  if (!expr || expr->op==XPP_ERROR)
  {
    str= malloc(2*8+strlen(op)+10);
    sprintf(str,"CO%s%sCO%s",Cnames[s1],op,Cnames[s2]);
    expr= xpp_run(xpp, str);
    if (!expr || expr->op==XPP_ERROR) 
    {
      // this will never happen. in case it does, I want a very weird
      // but consistent output
      expr= xpp_run(xpp, "0");
    }
  }

  sprintf(nb,"CO%s",Cnames[tgt]);
  unknown_t *U;
  U= calloc(1,sizeof(*U));
  U->target= strdup(nb);
  U->expr= expr;
  U->str= str;

  U->source= 'S';
  U->ptr= S;
  ptrbuf_put(B,U);
}

char *eval_screen_dim(Uscreen_t *S)
{
  ptrbuf_t *Bunk;
  strbuf_t *Bcode;
  size_t nunk; unknown_t **unk;
  char *code;
  int i;

  Bunk= ptrbuf_new(50);

  makeSref(Bunk, S, wLEF, wRIG, "-", wWID);
  makeSref(Bunk, S, wRIG, wLEF, "+", wWID);
  makeSref(Bunk, S, wTOP, wBOT, "-", wHEI);
  makeSref(Bunk, S, wBOT, wTOP, "+", wHEI);
  makeSref(Bunk, S, wWID, wRIG, "-", wLEF);
  makeSref(Bunk, S, wHEI, wBOT, "-", wTOP);

  ptrbuf_free(Bunk, (void***) &unk, &nunk);
  copy_fixed_knowns(nunk);

  Bcode= strbuf_init();
#if 0
  for(i=0;i<6;i++)  
  {
    char buf[50];
    sprintf(buf,"  int CO%s= -1;\n", Cnames[i]);
    strbuf_puts(Bcode, buf);
  }
#endif
  
  code= eval_assignments(Bcode,&nunk, unk);
  if (nunk)
  {
    Bcode= strbuf_init();
    strbuf_puts(Bcode, code);
    free(code);
    for(i=0;i<nunk;i++) 
    { 
      strbuf_printf(Bcode, "int %s= -1;\n", unk[i]->target);
      free_unknown(unk[i]);
    }
    code= strbuf_collect(Bcode);
  }
  if (unk) free(unk);
  return code;
}

char *eval_widget_layout(Uscreen_t *S, size_t *Rnunk, unknown_t ***Runk)
{
  get_widget_layout_unknowns(S,Runk,Rnunk);
  return eval_assignments(NULL,Rnunk, *Runk);
}
