#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include "expr.h"
#include "widget.h"
#ifndef KLASSES_NO_GFX
#include "gfx.h"
#endif
#include "bytecode.h"
#include "n-eval.h"

#ifdef KLASSES_NO_GFX
#define EDITFUNC(x) NULL
#else
#define EDITFUNC(x) x
#endif

#define BC_INT 1
#define BC_FLO 2
#define BC_STR 3
#define BC_END 4

typedef struct {
  int type; union { int Vint; float Vflo; char *Vstr; } d;
} bcarg_t;

static bcarg_t bcINT(int K)
{ 
  bcarg_t A; A.type= BC_INT;
  A.d.Vint= K; return A;
} 
static bcarg_t bcEND()
{
  bcarg_t A;
  A.type= BC_END;
  return A;
}

static bcarg_t bcFLO(float K)
{ 
  bcarg_t A; A.type= BC_FLO;
  A.d.Vflo= K; return A;
} 
static bcarg_t bcSTR(char *K)
{ 
  bcarg_t A; A.type= BC_STR;
  A.d.Vstr= K; return A;
} 



static void bcCALL(Uwidget_t *W,bc_buffer_t *bc,int func, ...)
{
  va_list ap;
  bcarg_t arg;
  bcarg_t *args;
  int N,i;

  args= NULL;
again:
  va_start(ap,func);
  for(N=0;;N++)
  {
    arg=va_arg(ap,bcarg_t);
    if (arg.type==BC_END) break;
    if (args) args[N]= arg;
  }
  va_end(ap);
  if (!args && N) { args= alloca(sizeof(arg)*N); goto again; }
  for(i=N-1;i>=0;i--)
  {
    arg= args[i];
    switch(arg.type)
    {
    case BC_INT: bc_buffer_put_int(bc, arg.d.Vint); break;
    case BC_FLO: bc_buffer_put_flo(bc, arg.d.Vflo); break;
    case BC_STR: bc_buffer_put_str(bc, arg.d.Vstr); break;
    }
  }
  bc_buffer_put_int(bc, func);
  bc_buffer_put_int(bc, W->global_number);
  bc_buffer_put(bc, GUII_CONSTRUCT);
}

static char *get_field(Uwidget_t *W, char *name, char *def)
{
  field_t *F;
  char *D;
  int i;
  Uklass_t *K;

  for(F=W->fields;F;F=F->next)
    if (!strcmp(F->name, name))
    {
      D= F->data;
      if (D && D[0]) return D;
      break;
    }

  K= W->klass;
  for(i=0;i<K->nfields;i++)
    if (!strcmp(K->fields[i].name,name))
    {
      D= K->fields[i].data;
      if (D && D[0]) return D;
      break;
    }

  return def;
}

static int get_field_int(Uwidget_t *W, char *name, int def)
{
  char *S;
  S= get_field(W, name, NULL);
  if (S) return strtol(S,NULL,0);
  return def;
}

static int get_field_chr(Uwidget_t *W, char *name, int def)
{
  char *S;
  S= get_field(W, name, NULL);
  if (S) return S[0];
  return def;
}

static float get_field_flo(Uwidget_t *W, char *name, float def)
{
  char *S;
  S= get_field(W, name, NULL);
  if (S) return strtof(S,NULL);
  return def;
}

static char *nustr(char *x)
{
  if (x) return x;
  else return "NULL";
}

static char *Cstr(char *x)
{
  char *R;
  int L;
  int i,j;
  char *legal= " abcdefghijklmnopqrstuvwxyz!@#$%^&*()_+=-[]{};:,./<>?`~"
               "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
  // char *hex= "0123456789abcdef";
  if (!x) return "NULL";
  L= strlen(x);
  R= malloc(4*L+2+1);
  j= 0;
  R[j++]= '"';
  for(i=0;x[i];i++)
   if (strchr(legal,x[i])) R[j++]= x[i];
   else { sprintf(R+j,"\\%03o", *(unsigned char*)(x+i)); j+= 4; }
  R[j++]= '"';
  R[j++]= 0;
  return R;
}

/***********************************************************************
                    WIDGET CLASSES BEGIN
 ***********************************************************************/

// ##W.BUTTON
static void Kconstruct_button(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W, B, GUIWC_BUTTON, 
                 bcSTR(get_field(W,"text",NULL)), 
                 bcINT(get_field_chr(W,"iconpos",'T')),
                 bcINT(0), // FIXME: this is the icon resource
                 bcINT(GFXCLASS_BUTTON), bcEND());
}

static field_t fields_button[]= 
 {
   { "text", TYP_STR, NULL },
   { "iconpos", TYP_INT, "T" },
   { "icon", TYP_COLORIMG, NULL },
 };
 
static Uklass_t klass_button= { 
   "Button", 3, fields_button, 
   NULL,
   EDITFUNC(ew_button), 
   Kconstruct_button 
};
// ##W.CANVAS
static void Kconstruct_canvas(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  int hs, vs;
  hs= get_field_chr(W,"horscroll", '1') == '1';
  vs= get_field_chr(W,"verscroll", '1' ) == '1';
  bcCALL(W,B, GUIWC_CANVAS, bcINT(hs), bcINT(vs), bcEND() );
}

static field_t fields_canvas[]=
{
  { "horscroll", TYP_INT, "1" },
  { "verscroll", TYP_INT, "1" },
};

static Uklass_t klass_canvas=
{
  "Canvas", 2, fields_canvas,
   NULL,
   EDITFUNC(ew_canvas),
   Kconstruct_canvas
};

// ##W.CHECKBOX
static void Kconstruct_checkbox(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W, B, GUIWC_CHECKBOX, bcSTR(get_field(W,"text","")), bcEND() );
}

static field_t fields_checkbox[]= 
 {
   { "text", TYP_STR, NULL },
 };

static Uklass_t klass_checkbox=
{
  "CheckBox", 1, fields_checkbox, 
   NULL, 
   EDITFUNC(ew_checkbox),
   Kconstruct_checkbox
};


// ##W.DIRLIST
static void Kconstruct_dirlist(Uscreen_t *S,Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_DIRLIST, bcEND() );
}

static Uklass_t klass_dirlist= {
   "DirList", 0, NULL,
   NULL, EDITFUNC(ew_none) , Kconstruct_dirlist};

// ##W.DROPDOWN
static void Kconstruct_dropdown(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_DROPDOWN, bcEND()); 
}

static Uklass_t klass_dropdown=
 { "DropDown", 0, NULL , NULL, EDITFUNC(ew_none), Kconstruct_dropdown };

// ##W.FILELIST
static void Kconstruct_filelist(Uscreen_t *S,Uwidget_t *W, bc_buffer_t *B)
{
  int ty, hi;
  ty= (get_field_chr(W, "types", '1')=='1');
  hi= (get_field_chr(W, "hidden", '1')=='1');
  bcCALL(W,B, GUIWC_FILELIST, bcINT(ty), bcINT(hi), bcEND() );
}

static field_t fields_filelist[]= 
 {
   { "hidden", TYP_INT, "1" },
   { "types", TYP_INT, "1" },
 };

static Uklass_t klass_filelist= {
   "FileList", 2, fields_filelist, 
   NULL, EDITFUNC(ew_filelist), Kconstruct_filelist };


// ##W.ICONLIST
static void Kconstruct_iconlist(Uscreen_t *S,Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_ICONLIST,
          bcINT(get_field_int(W, "fg",   0)),
          bcINT(get_field_int(W, "bg",   0xffffff)),
          bcINT(get_field_int(W, "vpad", 10)),
          bcINT(get_field_int(W, "hpad", 10)),
          bcINT(get_field_int(W, "max_text_width", 150)),
          bcINT(get_field_int(W, "multi_select", 0)), bcEND());
}

static field_t fields_iconlist[]= 
 {
  { "hpad", TYP_INT, "10" },
  { "vpad", TYP_INT, "10" },
  { "fg",   TYP_INT, "0x00" },
  { "bg",   TYP_INT, "0xffFFff" },
  { "max_text_width", TYP_INT, "150" },
  { "multi_select", TYP_INT, "0" }
 };

static Uklass_t klass_iconlist= {
   "IconList", 6, fields_iconlist,
   NULL, EDITFUNC(ew_iconlist), Kconstruct_iconlist };

// ##W.LABEL
static void Kconstruct_label(Uscreen_t *S,Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W, B, GUIWC_LABEL, 
                 bcINT(get_field_int(W,"font",1)),
                 bcINT(get_field_int(W,"fg", 0)),
                 bcINT(get_field_int(W,"bg", 0xededed)),
                 bcINT(get_field_chr(W,"just", 'C')),
                 bcSTR(get_field(W,"text","")), bcEND() );
}

static field_t fields_label[]=
 {
   { "text", TYP_STR, NULL },
   { "fg",   TYP_INT, "0x000000" },
   { "bg",   TYP_INT, "0xededed" },
   { "font", TYP_INT, "1" },
   { "just", TYP_INT, NULL },
 };

static Uklass_t klass_label=
{
   "Label", 5, fields_label, 
   NULL,
   EDITFUNC(ew_label),
   Kconstruct_label
};

// ##W.LIST
static void Kconstruct_list(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_LIST, bcEND());
}

static Uklass_t klass_list=
 { "List", 0, NULL , NULL, EDITFUNC(ew_none), Kconstruct_list, };

// ##W.LOGVIEW
static void Kconstruct_logview(Uscreen_t *S,Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_LOGVIEW, bcEND());
}

static Uklass_t klass_logview= {
   "LogView", 0, NULL, 
   NULL,
   EDITFUNC(ew_none),
   Kconstruct_logview };

// ##W.PANEL
static void Kconstruct_panel(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  int sa;
  int i,nc;
  Uwidget_t *cw;
  Uscreen_t *sub;

  sa= 0;  // size of array 
  bc_buffer_put_int(B,W->n_subs); sa++;
  for(i=0;i<W->n_subs;i++)
  {
    sub= W->subs[i];
    for(nc=0,cw=sub->widgets;cw;cw=cw->next,nc++)  ;
    bc_buffer_put_int(B, nc);
    sa++;
    for(cw=sub->widgets;cw;cw=cw->next)
    {
      bc_buffer_put_int(B, cw->global_number);
      sa++;
    }
  }
  bc_buffer_put_int(B, sa);
  bc_buffer_put(B, GUII_INTARRAY);
 
  bc_buffer_put_int(B, GUIWC_PANEL);
  bc_buffer_put_int(B, W->global_number);
  bc_buffer_put(B, GUII_CONSTRUCT);
}

static Uklass_t klass_panel =
{
  "Panel", 0, NULL, NULL, EDITFUNC(ew_panel), Kconstruct_panel,
};

// ##W.PIXGRID
static void Kconstruct_pixgrid(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_PIXGRID, bcEND());
}

static Uklass_t klass_pixgrid=
 { "PixGrid", 0, NULL, NULL, EDITFUNC(ew_none) , Kconstruct_pixgrid, };


// ##W.PROGRESSBAR
static void Kconstruct_progressbar(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_PROGRESSBAR, 
                bcINT(get_field_int(W,"font",1)),
                bcINT(get_field_int(W,"fg",0x00)),
                bcINT(get_field_int(W,"bg",0xededed)),
                bcINT(get_field_int(W,"fill", 0xcc)), bcEND() );
}

static field_t fields_progressbar[]=
 {
   { "fg",   TYP_INT, "0x000000" },
   { "bg",   TYP_INT, "0xededed" },
   { "fill", TYP_INT, "0x0000cc" },
   { "font", TYP_INT, "1" },
 };

static Uklass_t klass_progress=
{
   "ProgressBar", 4, fields_progressbar,
   NULL, 
   EDITFUNC(ew_progressbar),
   Kconstruct_progressbar,
};

// ##W.RADIO
static void Kconstruct_radio(Uscreen_t *S,Uwidget_t *W, bc_buffer_t *B)
{
  char *gr;
  Uwidget_t *ref;
  gr= get_field(W,"group",NULL);
  if (!gr)
     ref= W; 
  else
  {
     ref= find_widget_in_screen(S, gr);
     if (!ref) ref= W;
  }
  bcCALL(W, B, GUIWC_RADIO, 
                bcSTR(get_field(W,"text","")),
                bcINT(ref->global_number), bcEND());
}

static field_t fields_radio[]= 
 {
   { "text", TYP_STR, NULL },
   { "group", TYP_INT, "-1"},
 };

static Uklass_t klass_radio=
{
  "Radio", 2, fields_radio, 
  NULL, 
  EDITFUNC(ew_radio),
  Kconstruct_radio
};

// ##W.SCROLLBAR
static void Kconstruct_scrollbar(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_SCROLLBAR,
		bcINT(get_field_chr(W,"vertical",'0')!='0'),
                bcFLO(get_field_flo(W,"init_start",0.0)),
                bcFLO(get_field_flo(W,"init_end",1.0)),
                bcFLO(get_field_flo(W,"init_step",0.1)), bcEND() );
}

static field_t fields_scrollbar[]=
 {
   { "vertical", TYP_INT, "0" },
   { "init_start", TYP_INT, "0.0" },
   { "init_end", TYP_INT, "1.0" },
   { "init_step", TYP_INT, "0.1" },
 };

static Uklass_t klass_scrollbar=
{
  "ScrollBar", 4, fields_scrollbar, NULL, EDITFUNC(ew_scrollbar),
  Kconstruct_scrollbar
};

// ##W.SLIDER
static void Kconstruct_slider(Uscreen_t *S,Uwidget_t *W,bc_buffer_t* B)
{
  int orient;
  orient= (get_field_chr(W,"orientation",'0')=='0') ? 'H' : 'V';
  bcCALL(W,B, GUIWC_SLIDER, bcINT(orient), bcEND());
}

static field_t fields_slider[]=
{
   { "orientation", TYP_INT, "0" },
};
static Uklass_t klass_slider=
{
  "Slider", 1, fields_slider, NULL, EDITFUNC(ew_slider), Kconstruct_slider,
};

// ##W.SPINNER 
static void Kconstruct_spinner(Uscreen_t *S,Uwidget_t *W,bc_buffer_t* B)
{
  bcCALL(W,B, GUIWC_SPINNER, bcEND() );
}
static field_t fields_spinner[]= 
 {
   { "type", TYP_INT, "I" },
   { "decimals", TYP_INT, "2" },
 };

static Uklass_t klass_spinner= 
{
   "Spinner", 1, fields_spinner, NULL, EDITFUNC(ew_none),
   Kconstruct_spinner
};

// ##W.TEXTENTRY
static void Kconstruct_textentry(Uscreen_t *S, Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W,B, GUIWC_TEXTENTRY, bcSTR(get_field(W,"text","")), bcEND());
}

static field_t fields_textentry[]=
 {
   { "text", TYP_STR, NULL },
 };

static Uklass_t klass_textentry=
{
   "TextEntry", 1, fields_textentry, 
   NULL, 
   EDITFUNC(ew_textentry),
   Kconstruct_textentry,
};


// ##W.TOGGLE
static void Kconstruct_toggle(Uscreen_t *S,Uwidget_t *W, bc_buffer_t *B)
{
  bcCALL(W, B, GUIWC_BUTTON, 
                 bcSTR(get_field(W,"text",NULL)), 
                 bcINT(get_field_chr(W,"iconpos",'T')),
                 bcINT(0), // FIXME: this is the icon resource
                 bcINT(GFXCLASS_TOGGLE), bcEND());
}

static field_t fields_toggle[]=
 {
   { "text", TYP_STR, NULL },
   { "iconpos", TYP_INT, "T" },
   { "icon", TYP_COLORIMG, NULL },
 };

static Uklass_t klass_toggle= {
   "Toggle", 3, fields_toggle,
   NULL,
   EDITFUNC(ew_toggle),
   Kconstruct_toggle };


Uklass_t *klasses[]=
{ &klass_button,
  &klass_canvas,
  &klass_checkbox,
  &klass_dropdown,
  &klass_filelist,
  &klass_dirlist,
  &klass_iconlist,
  &klass_label,
  &klass_list,
  &klass_logview,
  &klass_panel,
  &klass_pixgrid,
  &klass_progress,
  &klass_radio,
  &klass_scrollbar,
  &klass_slider,
  &klass_spinner,
  &klass_textentry,
  &klass_toggle,
  NULL
};


#ifndef KLASSES_NO_GFX
Uklass_t *choose_klass()
{
  Uklass_t *ret;
  gfx_event_t event;
  ret= NULL;
  gfx_SCRwidsel();
  while(1)
   if (gfx_event(NULL,&event))
   switch(event.type)
   {
   case WwselCANCEL:
   case GFXEV_QUIT:                             goto done;
   case WwselBUTTON:    ret= &klass_button;     goto done;
   case WwselCANVAS:    ret= &klass_canvas;     goto done;
   case WwselCHECKBOX:  ret= &klass_checkbox;   goto done;
   case WwselDROPDOWN:  ret= &klass_dropdown;   goto done;
   case WwselFILELIST:  ret= &klass_filelist;   goto done;
   case WwselDIRLIST:   ret= &klass_dirlist;    goto done;
   case WwselICONLIST:  ret= &klass_iconlist;   goto done;
   case WwselLABEL:     ret= &klass_label;      goto done;
   case WwselLIST:      ret= &klass_list;       goto done;
   case WwselLOGVIEW:   ret= &klass_logview;    goto done;
   case WwselPANEL:     ret= &klass_panel;      goto done;
   case WwselPIXGRID:   ret= &klass_pixgrid;    goto done;
   case WwselPROGRESS:  ret= &klass_progress;   goto done;
   case WwselRADIO:     ret= &klass_radio;      goto done;
   case WwselSCROLLBAR: ret= &klass_scrollbar;  goto done;
   case WwselSLIDER:    ret= &klass_slider;     goto done;
   case WwselSPINNER:   ret= &klass_spinner;    goto done;
   case WwselTEXTENTRY: ret= &klass_textentry;  goto done;
   case WwselTOGGLE:    ret= &klass_toggle;     goto done;
   }
done:
  gfx_leave_screen();
  return ret;
}
#endif

