
#define WOMAP_PROTO static

typedef struct { void *root; } womap_t;

WOMAP_PROTO void* womap_find(womap_t *map,uint32_t key);
WOMAP_PROTO void* womap_insert(womap_t* map, uint32_t key, void *data);
WOMAP_PROTO void *womap_remove(womap_t* map, uint32_t key);
WOMAP_PROTO void *womap_destroy(womap_t *map);
WOMAP_PROTO womap_t  *womap_new();

typedef struct
{
  union { struct { void* child[2];           } inte;  
          struct { void *data; uint32_t key; } leaf; } d;
  unsigned internal:1, bit:5;
} womap_node_t;

 // careful, qumap also uses the same things
#ifndef Q_KEY
#define Q_KEY(X)    ((X)->d.leaf.key)
#define Q_DATA(X)   ((X)->d.leaf.data)
#define Q_CHLD(X,Y) ((X)->d.inte.child[Y])
#endif


static inline int womap_dir(uint32_t key,womap_node_t *node);
static inline womap_node_t* womap_leaf_node(uint32_t key, void *data);
static inline womap_node_t* womap_internal_node(uint32_t bit);
static uint8_t womap_crit_bit(uint32_t X, uint32_t Y);

static inline int womap_dir(uint32_t key,womap_node_t *node)
{
  return (key >> node->bit)&1;
}

static inline womap_node_t* womap_leaf_node(uint32_t key, void *data)
{
  womap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  Q_KEY(N)= key;
  Q_DATA(N)= data; 
  return N;
}

static inline womap_node_t* womap_internal_node(uint32_t bit)
{
  womap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->bit= bit;
  return N;
}

static uint8_t womap_crit_bit(uint32_t X, uint32_t Y)
{
  uint32_t i,j;
  static const uint32_t B[]=
  {
    0x55555555,
    0x33333333,
    0x0f0f0f0f,
    0x00ff00ff,
    0x0000ffff,
  };
  uint32_t A;
  A= X ^ Y;
  
  for(i=1;i<32;i<<=1) A |= (A<<i);
  A = ~A;
  for(j=0,i=1;i<32;i<<=1,j++) A= (A&B[j]) + ((A>>i)&B[j]);
  return A;
}


WOMAP_PROTO void* womap_find(womap_t *map,uint32_t key)
{
  womap_node_t *node;

  node= map->root;
  if (!node) return NULL;

  while(node->internal) node= Q_CHLD(node, womap_dir(key,node));

  if (Q_KEY(node)==key) return Q_DATA(node);
                          else return NULL;
}


WOMAP_PROTO void* womap_insert(womap_t* map, uint32_t key, void *data)
{
  womap_node_t *node_elt;

  if (!map->root)
  {
     map->root= womap_leaf_node(key,data);
     return 0;
  }

  node_elt= map->root;
  while(node_elt->internal)
    node_elt= Q_CHLD(node_elt, womap_dir(key,node_elt));

  
  uint32_t elt;
  int dir_elt;
  uint8_t bit;
  womap_node_t *node_key;

  elt= Q_KEY(node_elt);

  if (elt==key) return Q_DATA(node_elt);

  bit= womap_crit_bit(elt,key);

  dir_elt= (elt>>bit)&1;

  node_key= womap_internal_node(bit);
  Q_CHLD(node_key,1-dir_elt)= womap_leaf_node(key,data);

  womap_node_t *parent;
  womap_node_t *node;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->bit>bit) break;
     parent= node;
     node= Q_CHLD(node, womap_dir(key,node));
  }

  Q_CHLD(node_key,dir_elt)= node;
  if (parent) { if (Q_CHLD(parent,0)==node)  Q_CHLD(parent,0)= node_key;
                                       else  Q_CHLD(parent,1)= node_key; }
  else map->root= node_key;
  return 0;
}

WOMAP_PROTO void *womap_remove(womap_t* map, uint32_t key)
{
  void *retv;

  if (!map->root) return NULL;

   
  womap_node_t *node;
  womap_node_t *sibling;
  womap_node_t *parent;
  womap_node_t *gparent;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= womap_dir(key, node);
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1-dir);
    node= Q_CHLD(node, dir);
  }

  if (Q_KEY(node)!=key) return NULL;
  retv= Q_DATA(node);
  free(node);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }

  if (Q_CHLD(gparent,0)==parent) Q_CHLD(gparent,0)= sibling;
                            else Q_CHLD(gparent,1)= sibling;
  free(parent);

  return retv;
}

WOMAP_PROTO void *womap_destroy(womap_t *map)
{
  womap_node_t *node;
  womap_node_t *parent;
  womap_node_t *gparent;
  womap_node_t *sibling;
  void *retv;

  if (!map->root) return 0;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1);
    node= Q_CHLD(node,0);
  }
  retv= Q_DATA(node);
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (Q_CHLD(gparent,0)==parent)
    Q_CHLD(gparent,0)= sibling;
  else
    Q_CHLD(gparent,1)= sibling;

  free(parent);

  return retv;
}

WOMAP_PROTO womap_t  *womap_new()
{
  womap_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}
